﻿using System;
using System.ComponentModel;
using System.Linq;

namespace ViewCommandExample.ViewModel
{
	/// <summary>
	/// Provides common properties and methods supporting view model classes.
	/// </summary>
	abstract class ViewModelBase : INotifyPropertyChanged
	{
		#region Common view properties

		private string displayName;

		/// <summary>
		/// Gets or sets the user-friendly name of this object. Derived classes can set this
		/// property to a new value, or override it to determine the value on-demand.
		/// </summary>
		public virtual string DisplayName
		{
			get { return this.displayName; }
			set
			{
				if (value != this.displayName)
				{
					this.displayName = value;
					OnPropertyChanged("DisplayName");
				}
			}
		}

		#endregion Common view properties

		#region Property update helpers

		/// <summary>
		/// Checks whether the new property value has changed and updates the backing field.
		/// </summary>
		/// <typeparam name="T"></typeparam>
		/// <param name="value">New property value.</param>
		/// <param name="field">Backing field.</param>
		/// <param name="propertyNames">Names of the properties to notify updated.</param>
		/// <returns>true if the value has changed, false otherwise.</returns>
		protected bool CheckUpdate<T>(T value, ref T field, params string[] propertyNames)
		{
			if ((value == null) != (field == null) ||      // Exactly one is null
				(value != null && !value.Equals(field)))   // Neither is null and they're not equal
			{
				field = value;
				OnPropertyChanged(propertyNames);
				return true;
			}
			return false;
		}

		#endregion Property update helpers

		#region INotifyPropertyChanged Member

		/// <summary>
		/// Raised when a property on this object has a new value.
		/// </summary>
		public event PropertyChangedEventHandler PropertyChanged;

		/// <summary>
		/// Raises this object's PropertyChanged event once for each property.
		/// </summary>
		/// <param name="propertyNames">The properties that has a new value.</param>
		protected void OnPropertyChanged(params string[] propertyNames)
		{
			foreach (string propertyName in propertyNames)
			{
				OnPropertyChanged(propertyName);
			}
		}

		/// <summary>
		/// Raises this object's PropertyChanged event.
		/// </summary>
		/// <param name="propertyName">The property that has a new value.</param>
		protected void OnPropertyChanged(string propertyName)
		{
#if DEBUG
			if (!TypeDescriptor.GetProperties(this).OfType<PropertyDescriptor>().Any(d => d.Name == propertyName))
			{
				throw new NotImplementedException("Notifying a change of non-existing property " + this.GetType().Name + "." + propertyName);
			}
#endif

			var handler = this.PropertyChanged;
			if (handler != null)
			{
				handler(this, new PropertyChangedEventArgs(propertyName));
			}
		}

		#endregion INotifyPropertyChanged Member
	}
}
